from dataclasses import dataclass
from .util import read_json
from copy import deepcopy


@dataclass
class MouthCoordinates:
    """Data class for mouth image coordinate and transformation data"""

    x: float  # Distance (pxls) from top border of mouth img to top border of pose img.
    y: float  # Distance (pxls) from left border of mouth img to left border of pose img.
    scale_x: float  # Multiple by which the image should be scaled along x-axis (width)
    scale_y: float  # Multiple by which the image should be scaled along y-axis (height)
    flip_x: bool  # Image should be flipped horizontally (about the center y-axis)
    rotation: float  # Counter clockwise degrees that the image should be rotated


@dataclass
class Pose:
    """Data class for storing data for specific character poses."""

    image_files: dict  # Dictionary containing paths to variations of the pose image.
    mouth_coordinates: MouthCoordinates  # Mouth coordinates / transformations.


@dataclass
class Emotions:
    """Data class for storing emotion specific poses"""

    explain: list[Pose]  # List of poses for the explain emotion.
    happy: list[Pose]  # List os poses for the happy emotion.
    sad: list[Pose]  # List of poses for the sad emotion.
    angry: list[Pose]  # List of poses for the sad emotion.
    confused: list[Pose]  # List of poses for the sad emotion.
    rhetorical: list[Pose]  # List of poses for the sad emotion.


def get_assets() -> Emotions:
    """Loads pose data from json file and returns as a dictionary.

    Returns:
        dict: Pose data, including paths to images, emotion specific poses, and mouth coords.
    """
    pose_data = read_json(file="pose_data.json")["emotions"]

    emotions = {}
    for emotion in pose_data.keys():
        poses = []
        for i, _ in enumerate(pose_data[emotion]):
            images = deepcopy(pose_data[emotion][i]["image_files"])
            coords = deepcopy(pose_data[emotion][i]["mouth_coordinates"])
            pose = {
                "image_files": images,
                "mouth_coordinates": MouthCoordinates(**coords),
            }
            poses.append(Pose(**pose))
        emotions[emotion] = poses
    return Emotions(**emotions)



import os
import json
import random

from PIL import Image
import numpy as np
import cv2

from .util import read_json
from .dataloader import get_assets
from .lipsync import viseme_sequencer


class FrameSequence:
    def __init__(self):
        self.pose_files = []
        self.mouth_files = []
        self.pose_images = []
        self.mouth_images = []
        self.mouth_coords = []
        self.final_frames = []


class animate:
    """Animates a cartoon that is lip synced to provieded audio voiceover."""

    def __init__(self, audio_file: str, txt_file: str, video_path: str):
        self.sequence = FrameSequence()
        self.video_path = video_path
        self.assets = get_assets()
        self.fps = 24

        # Create sequence of mouth images
        self.viseme_sequence = viseme_sequencer(audio_file, txt_file)
        self.build_mouth_sequence()

        time_t = len(self.sequence.mouth_files) / 24
        print(f"Num Created: {len(self.sequence.mouth_files)}")
        print(f"Duration: {time_t}")

        self.build_pose_sequence()

        self.frame_size = self.get_frame_size()
        # Create the animation
        self.compile_animation()

    def build_pose_sequence(self):
        """Creates the sequence of pose images for the video"""
        seconds_per_pose = 6
        seconds_per_blink = 3

        frame_count = 0
        blink_count = 0
        emotion = self.random_emotion()
        pose = random.choice(emotion)
        # Add a character pose frame for every frame of a mouth
        while len(self.sequence.pose_files) <= len(self.sequence.mouth_files):
            added_frames = len(self.sequence.pose_files) - frame_count
            if added_frames > seconds_per_pose * self.fps:
                # Change pose if 6 seconds has passed
                emotion = self.random_emotion()
                pose = random.choice(emotion)
                frame_count = len(self.sequence.pose_files)

            # Add 1 second worth of pose files to pose frames sequencs
            pose_files = [pose.image_files["open"] for _ in range(self.fps)]
            mouth_coords = [pose.mouth_coordinates for _ in range(self.fps)]
            self.sequence.pose_files.extend(pose_files)
            self.sequence.mouth_coords.extend(mouth_coords)

            # Character should blink every x seconds
            frames_since_blink = len(self.sequence.pose_files) - blink_count
            if frames_since_blink > seconds_per_blink * self.fps:
                self.blink(pose=pose)
                blink_count = len(self.sequence.pose_files)
                seconds_per_blink = random.randint(2, 6)

            # Shorten list of poses if too many were made
            if len(self.sequence.pose_files) > len(self.sequence.mouth_files):
                self.sequence.pose_files = self.sequence.pose_files[
                    : len(self.sequence.mouth_files)
                ]
                self.sequence.mouth_coords = self.sequence.mouth_coords[
                    : len(self.sequence.mouth_files)
                ]
                break

        # Prepend absolute path to all pose images
        self.sequence.pose_files = [
            f"{os.path.dirname(__file__)}{file}" for file in self.sequence.pose_files
        ]

        # Create mouth PIL image for every frame, with image transformations based on pose
        for i, _ in enumerate(self.sequence.mouth_files):
            transformed_image = mouth_transformation(
                mouth_file=self.sequence.mouth_files[i],
                mouth_coord=self.sequence.mouth_coords[i],
            )
            self.sequence.mouth_images.append(transformed_image)
        return

    def blink(self, pose):
        """Generates an animation sequence for eye blinking in a specific pose

        Args:
            pose (Pose): A Pose object containing pose configuration data

        Returns:
            int: Returns the number of frames that were generated for talling total generated
        """
        frames = []
        blink_duration = 0.4
        subsequence_duration = blink_duration / 5
        num_frames = int(self.fps * subsequence_duration)

        open_frames = [pose.image_files["open"] for _ in range(num_frames)]
        mid_frames = [pose.image_files["middle"] for _ in range(num_frames)]
        shut_frames = [pose.image_files["shut"] for _ in range(num_frames)]
        open_wait = [pose.image_files["open"] for _ in range(int(self.fps * 1.5))]

        frames.extend(open_frames)
        frames.extend(mid_frames)
        frames.extend(shut_frames)
        frames.extend(mid_frames)
        frames.extend(open_frames)

        self.sequence.pose_files.extend(frames)

        mouth_coords = [pose.mouth_coordinates for _ in range(len(frames))]
        self.sequence.mouth_coords.extend(mouth_coords)

        return len(frames)

    def build_mouth_sequence(self):
        """Generates a sequence of mouth images for video"""
        # Add mouth images to mouth image file sequence
        for i, _ in enumerate(self.viseme_sequence):
            if self.viseme_sequence[i].visemes:
                self.sequence.mouth_files.extend(self.viseme_sequence[i].visemes)

        # Prepend absolute path to mouth images
        for i, _ in enumerate(self.sequence.mouth_files):
            file = self.sequence.mouth_files[i]
            new_file = f"{os.path.dirname(__file__)}/assets/visemes/positive/{file}"
            self.sequence.mouth_files[i] = new_file

    def random_emotion(self):
        """Generates a random emotion to use in sequence

        Returns:
            list[Pose]: List of poses from a random emotion
        """
        emotions_list = list(self.assets.__dict__.keys())
        emotion = random.choice(emotions_list)
        return getattr(self.assets, emotion)

    def get_frame_size(self):
        pose_image = cv2.imread(self.sequence.pose_files[0])
        height, width, _ = pose_image.shape
        return (width, height)

    def compile_animation(self):
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        video = cv2.VideoWriter(self.video_path, fourcc, self.fps, self.frame_size)

        for i, _ in enumerate(self.sequence.pose_files):
            frame = cv2.imread(self.sequence.pose_files[i])
            if self.sequence.mouth_files[i] is not None:
                final_frame = render_frame(
                    pose_img=frame,
                    mouth_img=self.sequence.mouth_images[i],
                    mouth_coord=self.sequence.mouth_coords[i],
                )
            else:
                final_frame = frame
            video.write(final_frame)
        video.release()


def mouth_transformation(mouth_file, mouth_coord) -> Image:
    """Transforms mouth image with scaling, flipping, and rotation.
        This transformation is applied because, the same mouth shape images
        are used for different pose images, but the size, angle, and position
        of a mouth image will depend on which pose image is being used.

    Args:
        mouth_path (str): .png file path pointing to mouth image
        transformation (np.array): image transformation data for mouth

    Returns:
        Image: PIL Image object of mouth image with applied transformations
    """
    mouth = Image.open(mouth_file)
    # Flip mouth horizontally if necessary
    if mouth_coord.flip_x is True:
        mouth = mouth.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
    # Scale mouth image if necessary
    if mouth_coord.scale_y != 1:
        og_width, og_height = mouth.size
        new_width = int(abs(og_width * mouth_coord.scale_x))
        new_height = int(og_height * mouth_coord.scale_y)
        try:
            mouth = mouth.resize(new_width, new_height, Image.Resampling.LANCZOS)
        except:
            pass
    # Apply image rotation if necessary
    if mouth_coord.rotation != 0:
        mouth = mouth.rotate(-mouth_coord.rotation, resample=Image.Resampling.BICUBIC)
    return mouth


def render_frame(pose_img: Image, mouth_img: Image, mouth_coord):
    pose_img = Image.fromarray(pose_img)
    mouth_width, mouth_height = mouth_img.size

    # Location in pose image where mouth / viseme image will be added
    paste_coordinates = (
        int(mouth_coord.x - (mouth_width / 2)),
        int(mouth_coord.y - (mouth_height / 2)),
    )

    # Paste the mouth image onto the face image at the specified coordinates
    pose_img.paste(im=mouth_img, box=paste_coordinates, mask=mouth_img)
    np_image = np.array(pose_img)

    # Convert BGR PIL image to RGB (if necessary)
    if np_image.shape[2] == 3:
        np_image = cv2.cvtColor(np_image, cv2.COLOR_RGB2BGR)
    return np_image



from forcealign import ForceAlign
from .util import read_json
from dataclasses import dataclass
from datetime import datetime
from typing import Union
import random

FPS = 24
# Viseme image for silence (i.e. closed mouth, not speaking)
SILENT_VISEME = "9.png"
SILENT_PHONEME = "PAUSE"
# ARPAbet phonemes to simplified phonemes mapping
PHONEMES = read_json("phonemes.json")
# Simplified phonemes to viseme-sequence mapping
VISEMES = read_json("visemes.json")


@dataclass
class WordViseme:
    word: Union[str, None]  # Word associated with viseme
    visemes: list[str]  # List of mouth shape images for viseme
    phonemes: list[str]  # The phoneme associated with the viseme
    time_start: datetime  # The time the word starts (seconds)
    time_end: datetime  # The time the word ends (seconds)
    duration: float  # total word duration from start to end (seconds)
    total_frames: int  # total number of frames in video for word


def viseme_sequencer(audio_file: str, txt_file: str) -> list[WordViseme]:
    """Converts and audio / txt file to force aligned viseme sequence

    Args:
        audio_file (str): Path to audio file of a person speaking english (.wav or .mp3)
        txt_file (str): Path to txt file trascript of audio recording

    Returns:
        list[WordViseme]: A list of force aligned WordViseme objects
    """
    # Provide path to audio_file and corresponding txt_file with audio transcript
    aligner = ForceAlign(audio_file=audio_file, txt_file=txt_file)

    # Runs forced alignment algorithm and returns alignment results
    words = aligner.inference()

    first_word = words[0]
    print(f"Time Start: {first_word.time_start}")
    last_word = words[-1]
    total_duration = last_word.time_end
    target_frames = int(total_duration * 24)
    print(f"Target Duration: {total_duration}")
    print(f"Target Frames: {target_frames}")

    viseme_sequence = []
    for word in words:
        phonemes = [phoneme_no_stress(phoneme) for phoneme in word.phonemes]
        images = [phoneme_to_viseme(phoneme=phoneme) for phoneme in phonemes]
        duration = word.time_end - word.time_start
        total_frames = int((duration / total_duration) * target_frames)
        if total_frames == 0 and random.choice([True, False]):
            total_frames = 1
        elif random.choice([True, False]):
            total_frames += 1

        # If viseme is more than one frame long
        visemes = generate_viseme_frames(sequence=images, total_frames=total_frames)
        total_frames = len(visemes)

        viseme_sequence.append(
            WordViseme(
                word=word.word,
                visemes=visemes,
                phonemes=phonemes,
                time_start=word.time_start,
                time_end=word.time_end,
                duration=duration,
                total_frames=total_frames,
            )
        )

    # Add silence viseme (closed mouth) between speaking visemes
    finished_sequence = []
    for i, _ in enumerate(viseme_sequence):
        finished_sequence.append(viseme_sequence[i])
        if i == len(viseme_sequence) - 1:
            break

        silent_viseme = get_silent_viseme(
            viseme_sequence[i], viseme_sequence[i + 1], total_duration, target_frames
        )
        if silent_viseme:
            finished_sequence.append(silent_viseme)

    return finished_sequence


def generate_viseme_frames(sequence: list, total_frames: int) -> list:
    """Generates the complete viseme frame sequence for word viseme

    Args:
        sequence (list): List of visemes in word
        total_frames (int): Total frames allocated to full word

    Returns:
        list: Completed viseme video sequence of images for word
    """
    frames_per_subviseme = total_frames // len(sequence)
    remainder_end = total_frames % len(sequence)
    if frames_per_subviseme == 0:
        if random.choice([True, False]):
            frames_per_subviseme = 1
        else:
            return []

    viseme_frames = []
    for i, _ in enumerate(sequence):
        if len(sequence[i]) > frames_per_subviseme:
            viseme_frames.extend(sequence[i][:frames_per_subviseme])
        elif len(sequence[i]) < frames_per_subviseme:
            seq = upsample(sequence[i], frames_per_subviseme)
            viseme_frames.extend(seq)
        else:
            viseme_frames.extend(sequence[i])

    # Upsample the frames to target length
    if len(viseme_frames) < total_frames:
        viseme_frames = upsample(viseme_frames, total_frames)
    return viseme_frames


def upsample(sequence, length):
    repetitions = length // len(sequence)
    remainder = length % len(sequence)
    upsampled = [elem for elem in sequence for _ in range(repetitions)]

    final_upsampled = []
    for i, _ in enumerate(upsampled):
        if i > (len(upsampled) - remainder - 1):
            final_upsampled.extend([upsampled[i], upsampled[i]])
        else:
            final_upsampled.append(upsampled[i])

    return final_upsampled


def phoneme_no_stress(phoneme: str) -> str:
    """Removes stress symbol from an ARPAbet phoneme

    Args:
        phoneme (str): An ARPAbet phoneme

    Returns:
        str: ARPAbet phoneme without stress symbol
    """
    if phoneme[-1].isdigit():
        return phoneme[:-1]
    else:
        return phoneme


def phoneme_to_viseme(phoneme: str) -> list[str]:
    """Converts a phoneme to a viseme image sequence

    Args:
        phoneme (str): An ARPAbet phoneme

    Returns:
        str: A list of images files for the viseme
    """
    phoneme = phoneme_no_stress(phoneme=phoneme)
    simplified_phone = PHONEMES[phoneme]
    viseme = VISEMES[simplified_phone]
    return viseme


def get_silent_viseme(current_viseme, next_viseme, total_duration, target_frames):
    # The time the silent viseme should start after previous viseme (i.e. the next frame)
    delta = 0.00000000000000000001

    # Get start time, end time, and total duration of silence
    silence_start = current_viseme.time_end + delta
    silence_end = next_viseme.time_start - delta
    duration = silence_end - silence_start

    # Get number of frames for silence segment (24 frames per second of silence)
    total_frames = int((duration / total_duration) * target_frames)
    if total_frames == 0 and random.choice([True, False]):
        total_frames += 1
    elif random.choice([True, False]):
        total_frames += 1

    # Create frames for silence
    silent_visemes = [SILENT_VISEME for _ in range(total_frames)]
    phonemes = [SILENT_PHONEME for _ in range(total_frames)]
    return WordViseme(
        word=None,
        visemes=silent_visemes,
        phonemes=phonemes,
        time_start=silence_start,
        time_end=silence_end,
        duration=duration,
        total_frames=total_frames,
    )



import json
import os
from scipy.io import wavfile
from scipy.signal import resample
import numpy as np


def read_json(file: str) -> dict:
    """Reads a json file to dictionary

    Args:
        path (str): Path to the json file.

    Returns:
        dict: Python dictionary with json data.
    """
    path = f"{os.path.dirname(__file__)}/assets/{file}"
    with open(path, "r") as file:
        data = json.load(file)
    return data


def write_json(data: dict, file: str) -> None:
    """Writes a Python dictionary to a json file.

    Args:
        data (dict): Python dictionary with data.
        path (str): The path to the output json file.
    """
    path = f"{os.path.dirname(__file__)}/assets/{file}"
    with open(path, "w") as file:
        json.dump(data, file, indent=4)


def resample_audio(
    audio_file: str,
    output_file: str = None,
    target_sr: int = 48000,
    padding: bool = True,
):
    """Resamples audio to a target sample rate

    Args:
        audio_file (str): Path to audio file to be resampled
        output_file (str, optional): Path to save new file. Defaults to None.
        target_sr (int, optional): Target sample rate. Defaults to 48000.
        padding (bool, optional): Extends duration of audio to nearest integer number of seconds.

    Returns:
        tuple: Returns numpy array of resampled audio and the new audio sample rate
    """
    original_sample_rate, audio_data = wavfile.read(audio_file)
    resample_ratio = target_sr / original_sample_rate
    total_samples = int(len(audio_data) * resample_ratio)

    # Resamples the audio to have only 'total_samples' number of samples
    resampled_audio = resample(x=audio_data, num=total_samples)
    resampled_audio = resampled_audio.astype(np.int16)

    # Adds silence to end of audio so num samples is divisible by sample rate
    zero_pad = target_sr - (len(resampled_audio) % target_sr)
    if padding and int(zero_pad) not in [0, target_sr]:
        silence = np.zeros(zero_pad, dtype=np.int16)
        resampled_audio = np.concatenate((resampled_audio, silence))

    if output_file:
        wavfile.write(output_file, target_sr, resampled_audio)

    padding_secs = round((padding / target_sr), 2)
    return resampled_audio, target_sr
